open Format
open Term
open Substitution
open Overlap
open Rewriting
open Yices

let geq_rule a b = geq (IVar a) (IVar b)

let gt_rule a b = gt (IVar a) (IVar b)
  
let rec k_steps k b ~rs_minus_cs ~cs ~below_s ~goal u =
  let c0 = 
    if k = 0 then
      Or []
    else
      let l =
	if List.mem b rs_minus_cs then
	  [ And [gt_rule b rule;
		 k_steps (k - 1) b ~rs_minus_cs ~cs ~below_s ~goal v]
          | rule <- rs_minus_cs; v <- Rewriting.rewrite [rule] u ]
	else
          []
      in
      if List.mem u below_s then
	Or (l @
	    [ k_steps (k - 1) b ~rs_minus_cs ~cs ~below_s ~goal v
            | v <- Rewriting.rewrite cs u ])
      else 
	Or l
  in
  Or (c0 :: [ c | (u', c) <- goal; u = u' ])

let separate l = 
  let us, cs = List.split l in
  (us, And cs)

let rec l_multistep a b rs_minus_cs cs = function
  | V _ as t -> [(t, And [])]
  | F (f, ts) as t -> 
     Listx.unique
       (l_multistep_root a b rs_minus_cs cs t @
	[ let (us, c) = separate l in 
	  (F (f, us), c)
	| l <- Listx.pi [ l_multistep a b rs_minus_cs cs t | t <- ts ] ])
and l_multistep_root a b rs_minus_cs cs t = 
  Listx.unique
    [ (substitute tau r,
       if List.mem (l, r) rs_minus_cs then
	 And [Or [if List.mem a rs_minus_cs then gt_rule  a (l, r) else Or [];
		  if List.mem b rs_minus_cs then geq_rule b (l, r) else Or []];
	      c]
       else
	 And [])
    | (l, r) <- rs_minus_cs @ cs;
      Some sigma <- [pattern_match l t];
      (tau, c) <- l_multistep_subst a b rs_minus_cs cs sigma ]
and l_multistep_subst a b rs_minus_cs cs sigma =
  [ separate l
  | l <- Listx.pi [ [ ((x, u), c) | (u, c) <- l_multistep a b rs_minus_cs cs t ] 
                  | (x, t) <- sigma ] ]
    
let decreasing_constraint1 k rs_minus_cs cs (t, a, p, s, b, u) =
  let below_s = Listx.diff (reachable_all cs [s]) [s] in 
  let goal : (Term.t * Rule.t Yices.t) list =
    l_multistep a b rs_minus_cs cs t @
    (if p = [] then
       [ (t', And []) | t' <- reachable_all cs [t] ]
     else
       [])
  in
  if (* List.mem u below_s && *)
     Listx.intersect [ v | (v, _) <- goal ] (reachable_all cs [u]) then
    And []
  else
    k_steps k b ~rs_minus_cs ~cs ~below_s ~goal u

let critical_peaks rs =
  [ (replace (substitute sigma l2) (substitute sigma r1) p,
     (l1, r1), p,
     substitute sigma l2,
     (l2, r2),
     substitute sigma r2)
  | ((l1, r1), p, (l2, r2), sigma) <- overlap rs ]
    
let decreasing_constraint k rs cs =
  let rs_minus_cs = Listx.diff rs cs in
  And ([ geq (IVar rule) (Int 0) | rule <- rs_minus_cs ] @
       [ decreasing_constraint1 k rs_minus_cs cs peak | peak <- critical_peaks rs ])

module I = Index.Make (Rule)
    
let hot ~yices ~termination_tool k rs cs =
  if Rules.left_linear rs && Sn.sn termination_tool cs [] = Result.YES then
    let rs_minus_cs = Listx.diff rs cs in
    match
      sat
	~tool:yices
	~int:rs_minus_cs
	~to_string:I.index
	(decreasing_constraint k rs cs)
    with
    | Some b -> Some [ (rule, List.assoc rule b.int) | rule <- rs_minus_cs ]
    | None -> None
  else 
    None
(*
let rs = Read.read_trs "4.trs"

let _ = hot ~yices:"yices" ~termination_tool:"natt -v:0" 4 rs [List.nth rs 0]

*)
